@@ -141,7 +141,7 @@ static char table_a2b_base64[] = {
 #define BASE64_PAD '='
 
 /* Max binary chunk size; limited only by available memory */
-#define BASE64_MAXBIN (PY_SSIZE_T_MAX/2 - sizeof(PyStringObject) - 3)
+#define BASE64_MAXBIN ((PY_SSIZE_T_MAX - 3) / 2)
 
 static unsigned char table_b2a_base64[] =
 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
@@ -196,7 +196,7 @@ binascii_a2b_uu(PyObject *self, PyObject *args)
     PyObject *rv;
     Py_ssize_t ascii_len, bin_len;
 
-    if ( !PyArg_ParseTuple(args, "s*:a2b_uu", &pascii) )
+    if ( !PyArg_ParseTuple(args, "y*:a2b_uu", &pascii) )
         return NULL;
     ascii_data = pascii.buf;
     ascii_len = pascii.len;
@@ -208,11 +208,11 @@ binascii_a2b_uu(PyObject *self, PyObject *args)
     ascii_len--;
 
     /* Allocate the buffer */
-    if ( (rv=PyString_FromStringAndSize(NULL, bin_len)) == NULL ) {
+    if ( (rv=PyBytes_FromStringAndSize(NULL, bin_len)) == NULL ) {
         PyBuffer_Release(&pascii);
         return NULL;
     }
-    bin_data = (unsigned char *)PyString_AS_STRING(rv);
+    bin_data = (unsigned char *)PyBytes_AS_STRING(rv);
 
     for( ; bin_len > 0 ; ascii_len--, ascii_data++ ) {
         /* XXX is it really best to add NULs if there's no more data */
@@ -282,7 +282,7 @@ binascii_b2a_uu(PyObject *self, PyObject *args)
     PyObject *rv;
     Py_ssize_t bin_len;
 
-    if ( !PyArg_ParseTuple(args, "s*:b2a_uu", &pbin) )
+    if ( !PyArg_ParseTuple(args, "y*:b2a_uu", &pbin) )
         return NULL;
     bin_data = pbin.buf;
     bin_len = pbin.len;
@@ -294,11 +294,11 @@ binascii_b2a_uu(PyObject *self, PyObject *args)
     }
 
     /* We're lazy and allocate to much (fixed up later) */
-    if ( (rv=PyString_FromStringAndSize(NULL, 2 + (bin_len+2)/3*4)) == NULL ) {
+    if ( (rv=PyBytes_FromStringAndSize(NULL, 2 + (bin_len+2)/3*4)) == NULL ) {
         PyBuffer_Release(&pbin);
         return NULL;
     }
-    ascii_data = (unsigned char *)PyString_AS_STRING(rv);
+    ascii_data = (unsigned char *)PyBytes_AS_STRING(rv);
 
     /* Store the length */
     *ascii_data++ = ' ' + (bin_len & 077);
@@ -320,10 +320,12 @@ binascii_b2a_uu(PyObject *self, PyObject *args)
     }
     *ascii_data++ = '\n';       /* Append a courtesy newline */
 
-    /* rv is cleared on error */
-    (void)_PyString_Resize(&rv,
+    if (_PyBytes_Resize(&rv,
                        (ascii_data -
-                        (unsigned char *)PyString_AS_STRING(rv)));
+                        (unsigned char *)PyBytes_AS_STRING(rv))) < 0) {
+        Py_DECREF(rv);
+        rv = NULL;
+    }
     PyBuffer_Release(&pbin);
     return rv;
 }
@@ -368,7 +370,7 @@ binascii_a2b_base64(PyObject *self, PyObject *args)
     Py_ssize_t ascii_len, bin_len;
     int quad_pos = 0;
 
-    if ( !PyArg_ParseTuple(args, "s*:a2b_base64", &pascii) )
+    if ( !PyArg_ParseTuple(args, "y*:a2b_base64", &pascii) )
         return NULL;
     ascii_data = pascii.buf;
     ascii_len = pascii.len;
@@ -383,11 +385,11 @@ binascii_a2b_base64(PyObject *self, PyObject *args)
     bin_len = ((ascii_len+3)/4)*3; /* Upper bound, corrected later */
 
     /* Allocate the buffer */
-    if ( (rv=PyString_FromStringAndSize(NULL, bin_len)) == NULL ) {
+    if ( (rv=PyBytes_FromStringAndSize(NULL, bin_len)) == NULL ) {
         PyBuffer_Release(&pascii);
         return NULL;
     }
-    bin_data = (unsigned char *)PyString_AS_STRING(rv);
+    bin_data = (unsigned char *)PyBytes_AS_STRING(rv);
     bin_len = 0;
 
     for( ; ascii_len > 0; ascii_len--, ascii_data++) {
@@ -447,15 +449,17 @@ binascii_a2b_base64(PyObject *self, PyObject *args)
 
     /* And set string size correctly. If the result string is empty
     ** (because the input was all invalid) return the shared empty
-    ** string instead; _PyString_Resize() won't do this for us.
+    ** string instead; _PyBytes_Resize() won't do this for us.
     */
     if (bin_len > 0) {
-        /* rv is cleared on error */
-        (void)_PyString_Resize(&rv, bin_len);
+        if (_PyBytes_Resize(&rv, bin_len) < 0) {
+            Py_DECREF(rv);
+            rv = NULL;
+        }
     }
     else {
         Py_DECREF(rv);
-        rv = PyString_FromStringAndSize("", 0);
+        rv = PyBytes_FromStringAndSize("", 0);
     }
     PyBuffer_Release(&pascii);
     return rv;
@@ -474,7 +478,7 @@ binascii_b2a_base64(PyObject *self, PyObject *args)
     PyObject *rv;
     Py_ssize_t bin_len;
 
-    if ( !PyArg_ParseTuple(args, "s*:b2a_base64", &pbuf) )
+    if ( !PyArg_ParseTuple(args, "y*:b2a_base64", &pbuf) )
         return NULL;
     bin_data = pbuf.buf;
     bin_len = pbuf.len;
@@ -490,11 +494,11 @@ binascii_b2a_base64(PyObject *self, PyObject *args)
     /* We're lazy and allocate too much (fixed up later).
        "+3" leaves room for up to two pad characters and a trailing
        newline.  Note that 'b' gets encoded as 'Yg==\n' (1 in, 5 out). */
-    if ( (rv=PyString_FromStringAndSize(NULL, bin_len*2 + 3)) == NULL ) {
+    if ( (rv=PyBytes_FromStringAndSize(NULL, bin_len*2 + 3)) == NULL ) {
         PyBuffer_Release(&pbuf);
         return NULL;
     }
-    ascii_data = (unsigned char *)PyString_AS_STRING(rv);
+    ascii_data = (unsigned char *)PyBytes_AS_STRING(rv);
 
     for( ; bin_len > 0 ; bin_len--, bin_data++ ) {
         /* Shift the data into our buffer */
@@ -518,10 +522,12 @@ binascii_b2a_base64(PyObject *self, PyObject *args)
     }
     *ascii_data++ = '\n';       /* Append a courtesy newline */
 
-    /* rv is cleared on error */
-    (void)_PyString_Resize(&rv,
+    if (_PyBytes_Resize(&rv,
                        (ascii_data -
-                        (unsigned char *)PyString_AS_STRING(rv)));
+                        (unsigned char *)PyBytes_AS_STRING(rv))) < 0) {
+        Py_DECREF(rv);
+        rv = NULL;
+    }
     PyBuffer_Release(&pbuf);
     return rv;
 }
@@ -540,7 +546,7 @@ binascii_a2b_hqx(PyObject *self, PyObject *args)
     Py_ssize_t len;
     int done = 0;
 
-    if ( !PyArg_ParseTuple(args, "s*:a2b_hqx", &pascii) )
+    if ( !PyArg_ParseTuple(args, "y*:a2b_hqx", &pascii) )
         return NULL;
     ascii_data = pascii.buf;
     len = pascii.len;
@@ -555,11 +561,11 @@ binascii_a2b_hqx(PyObject *self, PyObject *args)
     /* Allocate a string that is too big (fixed later)
        Add two to the initial length to prevent interning which
        would preclude subsequent resizing.  */
-    if ( (rv=PyString_FromStringAndSize(NULL, len+2)) == NULL ) {
+    if ( (rv=PyBytes_FromStringAndSize(NULL, len+2)) == NULL ) {
         PyBuffer_Release(&pascii);
         return NULL;
     }
-    bin_data = (unsigned char *)PyString_AS_STRING(rv);
+    bin_data = (unsigned char *)PyBytes_AS_STRING(rv);
 
     for( ; len > 0 ; len--, ascii_data++ ) {
         /* Get the byte and look it up */
@@ -595,10 +601,13 @@ binascii_a2b_hqx(PyObject *self, PyObject *args)
         Py_DECREF(rv);
         return NULL;
     }
-    /* rv is cleared on error */
-    if (_PyString_Resize(&rv,
+    if (_PyBytes_Resize(&rv,
                        (bin_data -
-                        (unsigned char *)PyString_AS_STRING(rv))) == 0) {
+                        (unsigned char *)PyBytes_AS_STRING(rv))) < 0) {
+        Py_DECREF(rv);
+        rv = NULL;
+    }
+    if (rv) {
         PyObject *rrv = Py_BuildValue("Oi", rv, done);
         PyBuffer_Release(&pascii);
         Py_DECREF(rv);
@@ -620,7 +629,7 @@ binascii_rlecode_hqx(PyObject *self, PyObject *args)
     unsigned char ch;
     Py_ssize_t in, inend, len;
 
-    if ( !PyArg_ParseTuple(args, "s*:rlecode_hqx", &pbuf) )
+    if ( !PyArg_ParseTuple(args, "y*:rlecode_hqx", &pbuf) )
         return NULL;
     in_data = pbuf.buf;
     len = pbuf.len;
@@ -633,11 +642,11 @@ binascii_rlecode_hqx(PyObject *self, PyObject *args)
     }
 
     /* Worst case: output is twice as big as input (fixed later) */
-    if ( (rv=PyString_FromStringAndSize(NULL, len*2+2)) == NULL ) {
+    if ( (rv=PyBytes_FromStringAndSize(NULL, len*2+2)) == NULL ) {
         PyBuffer_Release(&pbuf);
         return NULL;
     }
-    out_data = (unsigned char *)PyString_AS_STRING(rv);
+    out_data = (unsigned char *)PyBytes_AS_STRING(rv);
 
     for( in=0; in<len; in++) {
         ch = in_data[in];
@@ -655,7 +664,7 @@ binascii_rlecode_hqx(PyObject *self, PyObject *args)
                 /* More than 3 in a row. Output RLE. */
                 *out_data++ = ch;
                 *out_data++ = RUNCHAR;
-                *out_data++ = inend-in;
+                *out_data++ = (unsigned char) (inend-in);
                 in = inend-1;
             } else {
                 /* Less than 3. Output the byte itself */
@@ -663,10 +672,12 @@ binascii_rlecode_hqx(PyObject *self, PyObject *args)
             }
         }
     }
-    /* rv is cleared on error */
-    (void)_PyString_Resize(&rv,
+    if (_PyBytes_Resize(&rv,
                        (out_data -
-                        (unsigned char *)PyString_AS_STRING(rv)));
+                        (unsigned char *)PyBytes_AS_STRING(rv))) < 0) {
+        Py_DECREF(rv);
+        rv = NULL;
+    }
     PyBuffer_Release(&pbuf);
     return rv;
 }
@@ -684,7 +695,7 @@ binascii_b2a_hqx(PyObject *self, PyObject *args)
     PyObject *rv;
     Py_ssize_t len;
 
-    if ( !PyArg_ParseTuple(args, "s*:b2a_hqx", &pbin) )
+    if ( !PyArg_ParseTuple(args, "y*:b2a_hqx", &pbin) )
         return NULL;
     bin_data = pbin.buf;
     len = pbin.len;
@@ -697,11 +708,11 @@ binascii_b2a_hqx(PyObject *self, PyObject *args)
     }
 
     /* Allocate a buffer that is at least large enough */
-    if ( (rv=PyString_FromStringAndSize(NULL, len*2+2)) == NULL ) {
+    if ( (rv=PyBytes_FromStringAndSize(NULL, len*2+2)) == NULL ) {
         PyBuffer_Release(&pbin);
         return NULL;
     }
-    ascii_data = (unsigned char *)PyString_AS_STRING(rv);
+    ascii_data = (unsigned char *)PyBytes_AS_STRING(rv);
 
     for( ; len > 0 ; len--, bin_data++ ) {
         /* Shift into our buffer, and output any 6bits ready */
@@ -718,10 +729,12 @@ binascii_b2a_hqx(PyObject *self, PyObject *args)
         leftchar <<= (6-leftbits);
         *ascii_data++ = table_b2a_hqx[leftchar & 0x3f];
     }
-    /* rv is cleared on error */
-    (void)_PyString_Resize(&rv,
+    if (_PyBytes_Resize(&rv,
                        (ascii_data -
-                        (unsigned char *)PyString_AS_STRING(rv)));
+                        (unsigned char *)PyBytes_AS_STRING(rv))) < 0) {
+        Py_DECREF(rv);
+        rv = NULL;
+    }
     PyBuffer_Release(&pbin);
     return rv;
 }
@@ -737,7 +750,7 @@ binascii_rledecode_hqx(PyObject *self, PyObject *args)
     PyObject *rv;
     Py_ssize_t in_len, out_len, out_len_left;
 
-    if ( !PyArg_ParseTuple(args, "s*:rledecode_hqx", &pin) )
+    if ( !PyArg_ParseTuple(args, "y*:rledecode_hqx", &pin) )
         return NULL;
     in_data = pin.buf;
     in_len = pin.len;
@@ -747,7 +760,7 @@ binascii_rledecode_hqx(PyObject *self, PyObject *args)
     /* Empty string is a special case */
     if ( in_len == 0 ) {
         PyBuffer_Release(&pin);
-        return PyString_FromStringAndSize("", 0);
+        return PyBytes_FromStringAndSize("", 0);
     }
     else if (in_len > PY_SSIZE_T_MAX / 2) {
         PyBuffer_Release(&pin);
@@ -756,12 +769,12 @@ binascii_rledecode_hqx(PyObject *self, PyObject *args)
 
     /* Allocate a buffer of reasonable size. Resized when needed */
     out_len = in_len*2;
-    if ( (rv=PyString_FromStringAndSize(NULL, out_len)) == NULL ) {
+    if ( (rv=PyBytes_FromStringAndSize(NULL, out_len)) == NULL ) {
         PyBuffer_Release(&pin);
         return NULL;
     }
     out_len_left = out_len;
-    out_data = (unsigned char *)PyString_AS_STRING(rv);
+    out_data = (unsigned char *)PyBytes_AS_STRING(rv);
 
     /*
     ** We need two macros here to get/put bytes and handle
@@ -782,9 +795,9 @@ binascii_rledecode_hqx(PyObject *self, PyObject *args)
     do { \
              if ( --out_len_left < 0 ) { \
                       if ( out_len > PY_SSIZE_T_MAX / 2) return PyErr_NoMemory(); \
-                      if (_PyString_Resize(&rv, 2*out_len) < 0) \
-                        { PyBuffer_Release(&pin); return NULL; } \
-                      out_data = (unsigned char *)PyString_AS_STRING(rv) \
+                      if (_PyBytes_Resize(&rv, 2*out_len) < 0) \
+                        { Py_DECREF(rv); PyBuffer_Release(&pin); return NULL; } \
+                      out_data = (unsigned char *)PyBytes_AS_STRING(rv) \
                                                              + out_len; \
                       out_len_left = out_len-1; \
                       out_len = out_len * 2; \
@@ -833,16 +846,18 @@ binascii_rledecode_hqx(PyObject *self, PyObject *args)
             OUTBYTE(in_byte);
         }
     }
-    /* rv is cleared on error */
-    (void)_PyString_Resize(&rv,
+    if (_PyBytes_Resize(&rv,
                        (out_data -
-                        (unsigned char *)PyString_AS_STRING(rv)));
+                        (unsigned char *)PyBytes_AS_STRING(rv))) < 0) {
+        Py_DECREF(rv);
+        rv = NULL;
+    }
     PyBuffer_Release(&pin);
     return rv;
 }
 
 PyDoc_STRVAR(doc_crc_hqx,
-"(data, oldcrc) -> newcrc. Compute CRC-CCITT incrementally");
+"(data, oldcrc) -> newcrc. Compute hqx CRC incrementally");
 
 static PyObject *
 binascii_crc_hqx(PyObject *self, PyObject *args)
@@ -852,7 +867,7 @@ binascii_crc_hqx(PyObject *self, PyObject *args)
     unsigned int crc;
     Py_ssize_t len;
 
-    if ( !PyArg_ParseTuple(args, "s*i:crc_hqx", &pin, &crc) )
+    if ( !PyArg_ParseTuple(args, "y*i:crc_hqx", &pin, &crc) )
         return NULL;
     bin_data = pin.buf;
     len = pin.len;
@@ -879,17 +894,13 @@ binascii_crc32(PyObject *self, PyObject *args)
     Py_ssize_t len;
     int signed_val;
 
-    if (!PyArg_ParseTuple(args, "s*|I:crc32", &pbuf, &crc32val))
-        return NULL;
-    /* In Python 2.x we return a signed integer regardless of native platform
-     * long size (the 32bit unsigned long is treated as 32-bit signed and sign
-     * extended into a 64-bit long inside the integer object).  3.0 does the
-     * right thing and returns unsigned. http://bugs.python.org/issue1202 */
+    if (!PyArg_ParseTuple(args, "y*|I:crc32", &pbuf, &crc32val))
+    return NULL;
     buf = (Byte*)pbuf.buf;
     len = pbuf.len;
     signed_val = crc32(crc32val, buf, len);
     PyBuffer_Release(&pbuf);
-    return PyInt_FromLong(signed_val);
+    return PyLong_FromUnsignedLong(signed_val & 0xffffffffU);
 }
 #else  /* USE_ZLIB_CRC32 */
 /*  Crc - 32 BIT ANSI X3.66 CRC checksum files
@@ -1015,23 +1026,24 @@ binascii_crc32(PyObject *self, PyObject *args)
 { /* By Jim Ahlstrom; All rights transferred to CNRI */
     Py_buffer pbin;
     unsigned char *bin_data;
-    unsigned int crc = 0U;      /* initial value of CRC */
+    unsigned int crc = 0;       /* initial value of CRC */
     Py_ssize_t len;
-    int result;
+    unsigned int result;
 
-    if ( !PyArg_ParseTuple(args, "s*|I:crc32", &pbin, &crc) )
+    if ( !PyArg_ParseTuple(args, "y*|I:crc32", &pbin, &crc) )
         return NULL;
     bin_data = pbin.buf;
     len = pbin.len;
 
     crc = ~ crc;
-    while (len-- > 0)
-        crc = crc_32_tab[(crc ^ *bin_data++) & 0xffU] ^ (crc >> 8);
+    while (len-- > 0) {
+        crc = crc_32_tab[(crc ^ *bin_data++) & 0xff] ^ (crc >> 8);
         /* Note:  (crc >> 8) MUST zero fill on left */
+    }
 
-    result = (int)(crc ^ 0xFFFFFFFFU);
+    result = (crc ^ 0xFFFFFFFF);
     PyBuffer_Release(&pbin);
-    return PyInt_FromLong(result);
+    return PyLong_FromUnsignedLong(result & 0xffffffff);
 }
 #endif  /* USE_ZLIB_CRC32 */
 
@@ -1046,7 +1058,7 @@ binascii_hexlify(PyObject *self, PyObject *args)
     char* retbuf;
     Py_ssize_t i, j;
 
-    if (!PyArg_ParseTuple(args, "s*:b2a_hex", &parg))
+    if (!PyArg_ParseTuple(args, "y*:b2a_hex", &parg))
         return NULL;
     argbuf = parg.buf;
     arglen = parg.len;
@@ -1057,12 +1069,12 @@ binascii_hexlify(PyObject *self, PyObject *args)
         return PyErr_NoMemory();
     }
 
-    retval = PyString_FromStringAndSize(NULL, arglen*2);
+    retval = PyBytes_FromStringAndSize(NULL, arglen*2);
     if (!retval) {
         PyBuffer_Release(&parg);
         return NULL;
     }
-    retbuf = PyString_AS_STRING(retval);
+    retbuf = PyBytes_AS_STRING(retval);
 
     /* make hex version of string, taken from shamodule.c */
     for (i=j=0; i < arglen; i++) {
@@ -1087,7 +1099,7 @@ This function is also available as \"hexlify()\".");
 static int
 to_int(int c)
 {
-    if (isdigit(c))
+    if (Py_ISDIGIT(c))
         return c - '0';
     else {
         if (Py_ISUPPER(c))
@@ -1109,7 +1121,7 @@ binascii_unhexlify(PyObject *self, PyObject *args)
     char* retbuf;
     Py_ssize_t i, j;
 
-    if (!PyArg_ParseTuple(args, "s*:a2b_hex", &parg))
+    if (!PyArg_ParseTuple(args, "y*:a2b_hex", &parg))
         return NULL;
     argbuf = parg.buf;
     arglen = parg.len;
@@ -1122,22 +1134,22 @@ binascii_unhexlify(PyObject *self, PyObject *args)
      */
     if (arglen % 2) {
         PyBuffer_Release(&parg);
-        PyErr_SetString(PyExc_TypeError, "Odd-length string");
+        PyErr_SetString(Error, "Odd-length string");
         return NULL;
     }
 
-    retval = PyString_FromStringAndSize(NULL, (arglen/2));
+    retval = PyBytes_FromStringAndSize(NULL, (arglen/2));
     if (!retval) {
         PyBuffer_Release(&parg);
         return NULL;
     }
-    retbuf = PyString_AS_STRING(retval);
+    retbuf = PyBytes_AS_STRING(retval);
 
     for (i=j=0; i < arglen; i += 2) {
         int top = to_int(Py_CHARMASK(argbuf[i]));
         int bot = to_int(Py_CHARMASK(argbuf[i+1]));
         if (top == -1 || bot == -1) {
-            PyErr_SetString(PyExc_TypeError,
+            PyErr_SetString(Error,
                             "Non-hexadecimal digit found");
             goto finally;
         }
@@ -1187,7 +1199,7 @@ binascii_a2b_qp(PyObject *self, PyObject *args, PyObject *kwargs)
     static char *kwlist[] = {"data", "header", NULL};
     int header = 0;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s*|i", kwlist, &pdata,
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "y*|i", kwlist, &pdata,
           &header))
         return NULL;
     data = pdata.buf;
@@ -1222,8 +1234,7 @@ binascii_a2b_qp(PyObject *self, PyObject *args, PyObject *kwargs)
                 odata[out++] = '=';
                 in++;
             }
-            else if ((in + 1 < datalen) &&
-                     ((data[in] >= 'A' && data[in] <= 'F') ||
+            else if (((data[in] >= 'A' && data[in] <= 'F') ||
                       (data[in] >= 'a' && data[in] <= 'f') ||
                       (data[in] >= '0' && data[in] <= '9')) &&
                      ((data[in+1] >= 'A' && data[in+1] <= 'F') ||
@@ -1250,7 +1261,7 @@ binascii_a2b_qp(PyObject *self, PyObject *args, PyObject *kwargs)
             out++;
         }
     }
-    if ((rv = PyString_FromStringAndSize((char *)odata, out)) == NULL) {
+    if ((rv = PyBytes_FromStringAndSize((char *)odata, out)) == NULL) {
         PyBuffer_Release(&pdata);
         PyMem_Free(odata);
         return NULL;
@@ -1300,7 +1311,7 @@ binascii_b2a_qp (PyObject *self, PyObject *args, PyObject *kwargs)
     int crlf = 0;
     unsigned char *p;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s*|iii", kwlist, &pdata,
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "y*|iii", kwlist, &pdata,
           &quotetabs, &istext, &header))
         return NULL;
     data = pdata.buf;
@@ -1317,29 +1328,26 @@ binascii_b2a_qp (PyObject *self, PyObject *args, PyObject *kwargs)
     /* First, scan to see how many characters need to be encoded */
     in = 0;
     while (in < datalen) {
-        Py_ssize_t delta = 0;
         if ((data[in] > 126) ||
             (data[in] == '=') ||
             (header && data[in] == '_') ||
             ((data[in] == '.') && (linelen == 0) &&
-             (in + 1 == datalen || data[in+1] == '\n' ||
-              data[in+1] == '\r' || data[in+1] == 0)) ||
+             (data[in+1] == '\n' || data[in+1] == '\r' || data[in+1] == 0)) ||
             (!istext && ((data[in] == '\r') || (data[in] == '\n'))) ||
             ((data[in] == '\t' || data[in] == ' ') && (in + 1 == datalen)) ||
             ((data[in] < 33) &&
              (data[in] != '\r') && (data[in] != '\n') &&
-             (quotetabs ||
-            (!quotetabs && ((data[in] != '\t') && (data[in] != ' '))))))
+             (quotetabs || ((data[in] != '\t') && (data[in] != ' ')))))
         {
             if ((linelen + 3) >= MAXLINESIZE) {
                 linelen = 0;
                 if (crlf)
-                    delta += 3;
+                    odatalen += 3;
                 else
-                    delta += 2;
+                    odatalen += 2;
             }
             linelen += 3;
-            delta += 3;
+            odatalen += 3;
             in++;
         }
         else {
@@ -1351,11 +1359,11 @@ binascii_b2a_qp (PyObject *self, PyObject *args, PyObject *kwargs)
                 linelen = 0;
                 /* Protect against whitespace on end of line */
                 if (in && ((data[in-1] == ' ') || (data[in-1] == '\t')))
-                    delta += 2;
+                    odatalen += 2;
                 if (crlf)
-                    delta += 2;
+                    odatalen += 2;
                 else
-                    delta += 1;
+                    odatalen += 1;
                 if (data[in] == '\r')
                     in += 2;
                 else
@@ -1367,21 +1375,15 @@ binascii_b2a_qp (PyObject *self, PyObject *args, PyObject *kwargs)
                     (linelen + 1) >= MAXLINESIZE) {
                     linelen = 0;
                     if (crlf)
-                        delta += 3;
+                        odatalen += 3;
                     else
-                        delta += 2;
+                        odatalen += 2;
                 }
                 linelen++;
-                delta++;
+                odatalen++;
                 in++;
             }
         }
-        if (PY_SSIZE_T_MAX - delta < odatalen) {
-            PyBuffer_Release(&pdata);
-            PyErr_NoMemory();
-            return NULL;
-        }
-        odatalen += delta;
     }
 
     /* We allocate the output same size as input, this is overkill.
@@ -1402,13 +1404,13 @@ binascii_b2a_qp (PyObject *self, PyObject *args, PyObject *kwargs)
             (data[in] == '=') ||
             (header && data[in] == '_') ||
             ((data[in] == '.') && (linelen == 0) &&
-             (in + 1 == datalen || data[in+1] == '\n' ||
-              data[in+1] == '\r' || data[in+1] == 0)) ||
+             (data[in+1] == '\n' || data[in+1] == '\r' || data[in+1] == 0)) ||
             (!istext && ((data[in] == '\r') || (data[in] == '\n'))) ||
             ((data[in] == '\t' || data[in] == ' ') && (in + 1 == datalen)) ||
             ((data[in] < 33) &&
              (data[in] != '\r') && (data[in] != '\n') &&
-             (quotetabs || ((data[in] != '\t') && (data[in] != ' ')))))
+             (quotetabs ||
+            (!quotetabs && ((data[in] != '\t') && (data[in] != ' '))))))
         {
             if ((linelen + 3 )>= MAXLINESIZE) {
                 odata[out++] = '=';
@@ -1464,7 +1466,7 @@ binascii_b2a_qp (PyObject *self, PyObject *args, PyObject *kwargs)
             }
         }
     }
-    if ((rv = PyString_FromStringAndSize((char *)odata, out)) == NULL) {
+    if ((rv = PyBytes_FromStringAndSize((char *)odata, out)) == NULL) {
         PyBuffer_Release(&pdata);
         PyMem_Free(odata);
         return NULL;
@@ -1500,26 +1502,41 @@ static struct PyMethodDef binascii_module_methods[] = {
 };
 
 
-/* Initialization function for the module (*must* be called initbinascii) */
+/* Initialization function for the module (*must* be called PyInit_binascii) */
 PyDoc_STRVAR(doc_binascii, "Conversion between binary data and ASCII");
 
+
+static struct PyModuleDef binasciimodule = {
+    PyModuleDef_HEAD_INIT,
+    "binascii",
+    doc_binascii,
+    -1,
+    binascii_module_methods,
+    NULL,
+    NULL,
+    NULL,
+    NULL
+};
+
 PyMODINIT_FUNC
-initbinascii(void)
+PyInit_binascii(void)
 {
-    PyObject *m, *d, *x;
+    PyObject *m, *d;
 
     /* Create the module and add the functions */
-    m = Py_InitModule("binascii", binascii_module_methods);
+    m = PyModule_Create(&binasciimodule);
     if (m == NULL)
-        return;
+        return NULL;
 
     d = PyModule_GetDict(m);
-    x = PyString_FromString(doc_binascii);
-    PyDict_SetItemString(d, "__doc__", x);
-    Py_XDECREF(x);
 
-    Error = PyErr_NewException("binascii.Error", NULL, NULL);
+    Error = PyErr_NewException("binascii.Error", PyExc_ValueError, NULL);
     PyDict_SetItemString(d, "Error", Error);
     Incomplete = PyErr_NewException("binascii.Incomplete", NULL, NULL);
     PyDict_SetItemString(d, "Incomplete", Incomplete);
+    if (PyErr_Occurred()) {
+        Py_DECREF(m);
+        m = NULL;
+    }
+    return m;
 }
